from typing import List, Tuple

from i8_terminal.types.auto_complete_choice import AutoCompleteChoice
from i8_terminal.types.metric_param_type import MetricParamType
from i8_terminal.types.period_param_type import PeriodParamType


class MetricIdentifierParamType(AutoCompleteChoice):
    name = "metricidentifier"

    def __init__(self) -> None:
        self._metric_auto_comp = MetricParamType()
        self._period_auto_comp = PeriodParamType()

    def get_suggestions(
        self, keyword: str, pre_populate: bool = True, param_type: str = None, metric: str = None
    ) -> List[Tuple[str, str]]:
        if param_type == "metric":
            return self._metric_auto_comp.get_suggestions(keyword)
        else:
            return self._period_auto_comp.get_suggestions(keyword, metric=metric)  # type: ignore

    def __repr__(self) -> str:
        return "METRICIDENTIFIER"



import webbrowser

import click
import investor8_sdk

from i8_terminal.commands.user import user, webserver
from i8_terminal.common.cli import pass_command
from i8_terminal.config import APP_SETTINGS, save_user_settings


def get_login_authentication_request_id() -> str:
    resp = investor8_sdk.UserApi().create_login_authentication_request(body={"ReqType": 3})
    return str(resp.request_id)


def open_browser(request_id: str) -> None:
    url = f"https://www.investoreight.com/account/authorize?reqId={request_id}&redirectUrl=http://localhost:{APP_SETTINGS['app']['port']}"  # noqa: E501
    webbrowser.open(url)


def login_within_terminal() -> None:
    email = click.prompt("Email", show_default=False, type=str)
    password = click.prompt("Password", hide_input=True, show_default=False, type=str)
    body = {"Email": email, "Password": password}
    api_response = None
    try:
        api_response = investor8_sdk.UserApi().login_user(body=body)
        user_setting = {
            "user_id": api_response.user_id,
            "i8_core_token": api_response.token,
            "i8_core_api_key": api_response.api_key,
        }
        save_user_settings(user_setting)
        investor8_sdk.ApiClient().configuration.api_key["apiKey"] = api_response.api_key
        investor8_sdk.ApiClient().configuration.api_key["Authorization"] = api_response.token
        investor8_sdk.ApiClient().configuration.api_key_prefix["Authorization"] = "Bearer"
        click.echo("User logged in successfully!")
    except Exception as e:
        click.echo(e)


@user.command()
@click.option(
    "--terminal",
    is_flag=True,
    default=False,
    help="Login with your credentials in the console (Only applicable if you have a local Investoreight account).",
)
@pass_command
def login(terminal: bool) -> None:
    if terminal:
        login_within_terminal()
    else:
        request_id = get_login_authentication_request_id()
        open_browser(request_id)
        webserver.run_server(request_id)



from typing import Dict, List, Tuple

from i8_terminal.types.auto_complete_choice import AutoCompleteChoice


def get_value_fields() -> List[Tuple[str, str]]:
    return [
        ("value_numeric", "Absolute Value Numeric"),
        ("value", "Absolute Value"),
        ("overall_rank", "Overall Rank"),
        ("spx_rank", "Spx Rank"),
        ("dow_rank", "Dow Rank"),
        ("sector_rank", "Sector Rank"),
        ("industry_rank", "Industry Rank"),
        ("overall_percentile", "Overall Percentile"),
        ("spx_percentile", "Spx Percentile"),
        ("nasdaq_percentile", "Nasdaq Percentile"),
        ("dow_percentile", "Dow Percentile"),
        ("sector_percentile", "Sector Percentile"),
        ("industry_percentile", "Industry Percentile"),
    ]


class ScreeningValueFieldParamType(AutoCompleteChoice):
    name = "screeningvaluefield"
    metrics_default_pt: Dict[str, str] = {}

    def get_suggestions(self, keyword: str, pre_populate: bool = True, metric: str = None) -> List[Tuple[str, str]]:
        if not self.is_loaded:
            self.set_choices(get_value_fields())

        if pre_populate and keyword.strip() == "":
            return self._choices[: self.size]
        return self.search_keyword(keyword)

    def __repr__(self) -> str:
        return "SCREENINGVALUEFIELD"



import json
from typing import Dict, List, Tuple

import numpy as np

from i8_terminal.common.formatting import format_number_v2
from i8_terminal.common.metrics import get_all_metrics_df
from i8_terminal.types.auto_complete_choice import AutoCompleteChoice

PERIOD_TYPES: Dict[str, str] = {
    "fy": "mry",
    "q": "mrq",
    "ttm": "ttm",
    "ytd": "ytd",
    "d": "d",
    "r": "r",
    "mry": "mry",
    "mrq": "mrq",
}


def get_metrics_conditions_dict() -> Dict[str, str]:
    df = get_all_metrics_df()[["metric_name", "screening_bounds"]]
    return dict([(i, j) for i, j in zip(df.metric_name, df.screening_bounds)])


def get_metrics_default_period_types_dict() -> Dict[str, str]:
    df = get_all_metrics_df()[["metric_name", "period_type_default"]]
    return dict([(i, j) for i, j in zip(df.metric_name, df.period_type_default)])


def get_metrics_data_format_dict() -> Dict[str, str]:
    df = get_all_metrics_df()[["metric_name", "data_format"]]
    return dict([(i, j) for i, j in zip(df.metric_name, df.data_format)])


def get_metrics_screening_categories_dict() -> Dict[str, str]:
    df = get_all_metrics_df()[["metric_name", "screening_categories"]]
    return dict([(i, j) for i, j in zip(df.metric_name, df.screening_categories)])


class ConditionParamType(AutoCompleteChoice):
    name = "condition"

    def get_suggestions(
        self, keyword: str, pre_populate: bool = True, metric: str = None, period: str = None, value_field: str = None
    ) -> List[Tuple[str, str]]:
        if not self.is_loaded:
            self.metrics_conditions = get_metrics_conditions_dict()
            self.metrics_default_period_types = get_metrics_default_period_types_dict()
            self.metrics_data_format = get_metrics_data_format_dict()
            self.metrics_screening_categories = get_metrics_screening_categories_dict()
        metric_screening_bounds_dict = json.loads(
            self.metrics_conditions.get(metric, "").replace("'", '"')  # type: ignore
        )
        if value_field in ["overall_rank", "spx_rank", "dow_rank", "sector_rank", "industry_rank"]:
            self.set_choices([("5", ""), ("10", ""), ("20", ""), ("50", ""), ("100", ""), ("500", ""), ("1000", "")])
        elif value_field in [
            "overall_percentile",
            "nasdaq_percentile",
            "spx_percentile",
            "sector_percentile",
            "industry_percentile",
            "dow_percentile",
        ]:
            self.set_choices([("1", ""), ("3", ""), ("5", ""), ("10", ""), ("20", ""), ("50", "")])
        else:
            if self.metrics_data_format.get(metric, "") == "categorical":  # type: ignore
                metric_screening_categories = json.loads(
                    self.metrics_screening_categories.get(metric, "").replace("'", '"')  # type: ignore
                )
                self.set_choices(
                    [
                        (
                            c.get("category_name", ""),
                            c.get("category_display_name", ""),
                        )
                        for c in metric_screening_categories
                    ]
                )
            else:
                default_p_type = self.metrics_default_period_types.get(metric, "Q")  # type: ignore
                self.set_choices(
                    [
                        (
                            str(c),
                            format_number_v2(c, percision=1 if c < 1000 else 0, humanize=False if c < 1000 else True),
                        )  # type: ignore
                        for c in metric_screening_bounds_dict.get(
                            PERIOD_TYPES.get(period, "mrq")
                            if period
                            else PERIOD_TYPES.get(
                                default_p_type.lower() if default_p_type is not np.nan else "Q"  # type: ignore
                            ),
                            "",
                        )
                    ]
                )

        if pre_populate and keyword.strip() == "":
            return self._choices[: self.size]
        return self.search_keyword(keyword)

    def __repr__(self) -> str:
        return "CONDITION"



import shlex
from typing import Callable, List, Union, cast

import click
from click import Group
from click.decorators import F
from click.shell_completion import _resolve_context
from prompt_toolkit.document import Document


class CompleterContext:
    def __init__(
        self,
        cli: Callable[[F], Group],
        click_ctx: click.Context,
        tokens: List[str],
        used_options: List[str],
        last_option: Union[str, None],
        incomplete: str,
    ) -> None:
        self.cli = cli
        self.click_ctx = click_ctx
        self.tockens = tokens
        self.used_options = used_options
        self.last_option = last_option
        self.incomplete = incomplete


class CommandParser:
    def __init__(self, cli: Callable[[F], Group]) -> None:
        self.cli = cast(Group, cli)

    def parse(self, document: Document) -> Union[CompleterContext, None]:
        tokens = document.text.split(" ")
        used_options = [p for p in tokens if p.startswith("-")]
        last_option = tokens[-2] if len(tokens) > 2 and tokens[-2].startswith("-") else None

        try:
            args = shlex.split(document.text_before_cursor)
        except ValueError:
            # Invalid command, perhaps caused by missing closing quotation.
            return None

        cursor_within_command = document.text_before_cursor.rstrip() == document.text_before_cursor

        if args and cursor_within_command:
            # We've entered some text and no space, give completions for the
            # current word.
            incomplete = args.pop()
        else:
            # We've not entered anything, either at all or for the current
            # command, so give all relevant completions for this context.
            incomplete = ""
        ctx = _resolve_context(self.cli, {}, "", args)

        return CompleterContext(self.cli, ctx, tokens, used_options, last_option, incomplete)



import sys
from functools import update_wrapper
from typing import Any, Dict, Optional

import click
import investor8_sdk

from i8_terminal.config import USER_SETTINGS
from i8_terminal.utils_setup import get_version


def get_click_command_path(ctx: Any, parsed_options_dict: Optional[Dict[str, str]] = None) -> str:
    command_path = ctx.command_path.replace("  ", " ")
    cm_path = f"i8 {' '.join(command_path.split(' ')[1:])}"
    params = ctx.params
    args = []
    options = {}
    for p in ctx.command.params:
        if p.param_type_name == "argument":
            args.append(params[p.name])
        elif p.param_type_name == "option" and params[p.name] is not None:
            options[f"--{p.name}"] = params[p.name] if type(params[p.name]) != bool else None

    if parsed_options_dict:
        options = {**options, **parsed_options_dict}
    return (
        f"{cm_path} {' '.join([f'{k} {val}' if val else k for (k, val) in options.items()])} {''.join(args)}".rstrip()
    )


def log_terminal_usage(ctx: click.Context, exception: Optional[str] = "") -> None:
    if USER_SETTINGS.get("allow_terminal_logging"):
        investor8_sdk.UserApi().log_terminal_usage(
            body={
                "Command": ctx.obj["command"],
                "Version": get_version(),
                "OS": sys.platform,
                "AppInstanceId": USER_SETTINGS.get("app_instance_id"),
                "Exception": exception,
            }
        )


def pass_command(f: Any) -> Any:
    @click.pass_context
    def new_func(ctx: click.Context, *args: Any, **kwargs: Any) -> Any:
        ctx.obj["command"] = get_click_command_path(ctx)
        return ctx.invoke(f, *args, **kwargs)

    return update_wrapper(new_func, f)



from typing import Dict, List

from i8_terminal.common.metrics import get_all_metrics_df
from i8_terminal.i8_exception import I8Exception
from i8_terminal.service_result.column_info import ColumnInfo


class ColumnsContext:
    def __init__(self, col_infos: List[ColumnInfo]):
        self._col_infos = col_infos
        self._enrich_col_infos()
        self._col_info_dict = {ci.name: ci for ci in self._col_infos}

    def _enrich_col_infos(self) -> None:
        all_metrics_dict = self.get_metrics_dict()

        for ci in self._col_infos:
            if ci.col_type == "metric":
                if ci.name not in all_metrics_dict:
                    raise I8Exception(
                        f"Metric `{ci.name}` is not a known metric! You need to explicitly define the metadata!"
                    )
                ci.enrich(all_metrics_dict[ci.name])

    def get_metrics_dict(self) -> Dict[str, ColumnInfo]:
        metrics_df = get_all_metrics_df()
        metrics_dict: Dict[str, ColumnInfo] = {}
        for _, r in metrics_df.iterrows():
            metrics_dict[r["metric_name"]] = ColumnInfo(
                r["metric_name"], "metric", r["display_name"], r["data_format"], r["unit"]  # , r["colorable"]
            )
        return metrics_dict

    def get_col_infos(self) -> List[ColumnInfo]:
        return self._col_infos

    def get_col_info_dict(self) -> Dict[str, ColumnInfo]:
        return self._col_info_dict

    def get_col_info(self, name: str) -> ColumnInfo:
        if name not in self._col_info_dict:
            raise I8Exception(f"Column `{name}` is not found!")
        return self._col_info_dict[name]



from typing import Optional

import click
from rich.console import Console

import i8_terminal.api.earnings as earnings_api
from i8_terminal.commands.earnings import earnings
from i8_terminal.common.cli import pass_command
from i8_terminal.common.layout import df2Table
from i8_terminal.common.stock_info import validate_ticker
from i8_terminal.common.utils import export_data, export_to_html
from i8_terminal.config import APP_SETTINGS
from i8_terminal.service_result.earning_list_result import EarningsListResult
from i8_terminal.types.ticker_param_type import TickerParamType


@earnings.command()
@click.option("--ticker", "-k", type=TickerParamType(), required=True, callback=validate_ticker, help="Company ticker.")
@click.option("--export", "export_path", "-e", help="Filename to export the output to.")
@pass_command
def list(ticker: str, export_path: Optional[str]) -> None:
    """
    Lists upcoming company earnings.

    Examples:

    `i8 earnings list --ticker AAPL`

    """
    earnings_list: EarningsListResult = earnings_api.list(ticker, 10)
    if export_path:
        if export_path.split(".")[-1] == "html":
            df = earnings_list.to_df()
            table = df2Table(df)
            export_to_html(table, export_path)
        else:
            export_data(
                earnings_list.to_df("raw"),
                export_path,
                column_width=18,
                column_format=APP_SETTINGS["styles"]["xlsx"]["financials"]["column"],
            )
    else:
        df = earnings_list.to_df()
        console = Console()
        console.print(earnings_list._to_rich_table("humanize", "default"))



import os

import click

from i8_terminal.commands.user import user
from i8_terminal.common.cli import pass_command
from i8_terminal.config import USER_SETTINGS, restore_user_settings


@user.command()
@pass_command
def logout() -> None:
    if USER_SETTINGS:
        restore_user_settings()
        click.echo("✅ User logged out successfully!")
        os._exit(0)
    else:
        click.echo("You are already logged out!")



from datetime import datetime
from typing import Any, Dict, List, Optional

import arrow
import click
import investor8_sdk
import numpy as np
import pandas as pd
import plotly.express as px
import plotly.graph_objects as go
from click.types import DateTime
from pandas import DataFrame
from plotly.subplots import make_subplots
from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from rich.tree import Tree

from i8_terminal.app.layout import get_plot_default_layout
from i8_terminal.app.plot_server import serve_plot
from i8_terminal.commands.metrics import metrics
from i8_terminal.common.cli import get_click_command_path, pass_command
from i8_terminal.common.layout import df2Table, format_metrics_df
from i8_terminal.common.metrics import get_all_metrics_type_and_data_types_df
from i8_terminal.common.stock_info import validate_tickers
from i8_terminal.common.utils import PlotType, reverse_period
from i8_terminal.config import get_table_style
from i8_terminal.types.chart_param_type import ChartParamType
from i8_terminal.types.metric_param_type import MetricParamType
from i8_terminal.types.output_param_type import OutputParamType
from i8_terminal.types.period_type_param_type import PeriodTypeParamType
from i8_terminal.types.ticker_param_type import TickerParamType


def get_historical_metrics_df(
    tickers: List[str],
    metrics: List[str],
    period_type: Optional[str],
    from_date: Optional[datetime],
    to_date: Optional[datetime],
) -> DataFrame:
    if period_type:
        metrics = [f"{metric}.{period_type}" for metric in metrics]
    if from_date:
        historical_metrics = investor8_sdk.MetricsApi().get_historical_metrics(
            symbols=",".join(tickers),
            metrics=",".join(metrics),
            from_date=from_date.strftime("%Y-%m-%d"),
            to_date=to_date.strftime("%Y-%m-%d") if to_date else arrow.now().datetime.strftime("%Y-%m-%d"),
        )
    elif to_date and not from_date:
        historical_metrics = investor8_sdk.MetricsApi().get_historical_metrics(
            symbols=",".join(tickers),
            metrics=",".join(metrics),
            from_period_offset=-10,
            to_period_offset=0,
            to_date=to_date.strftime("%Y-%m-%d"),
        )
    elif not to_date and not from_date:
        historical_metrics = investor8_sdk.MetricsApi().get_historical_metrics(
            symbols=",".join(tickers),
            metrics=",".join(metrics),
            from_period_offset=-10,
            to_period_offset=0,
        )
    df = pd.DataFrame.from_records(
        [
            (ticker, metric, period_value.period, period_value.period_date_time, period_value.value)
            for ticker, metric_dict in historical_metrics.data.items()
            for metric, period_value_list in metric_dict.items()
            for period_value in period_value_list
        ],
        columns=["Ticker", "metric_name", "Period", "PeriodDateTime", "Value"],
    )
    metadata_df = pd.DataFrame([h.to_dict() for h in historical_metrics.metadata])
    df = pd.merge(df, metadata_df, on="metric_name")
    df[["data_format", "display_format"]] = df[["data_format", "display_format"]].replace("string", "str")
    df.rename(columns={"display_name": "Metric", "Value": "value"}, inplace=True)
    df["value"] = df.apply(lambda metric: data_format_mapper(metric), axis=1)
    return df


def data_format_mapper(metric: pd.Series) -> Any:
    if metric["data_format"] in ["int", "unsigned_int"]:
        return int(float(metric["value"]))
    elif metric["data_format"] == "float":
        return float(metric["value"])
    else:
        # Includes "datetime", "string" and "str"
        return str(metric["value"])


def create_fig(
    df: DataFrame, cmd_context: Dict[str, Any], tickers: List[str], chart_type: str, metrics_type_df: DataFrame
) -> go.Figure:
    vertical_spacing = 0.02
    layout = dict(
        title=cmd_context["plot_title"],
        autosize=True,
        hovermode="closest",
        legend=dict(font=dict(size=11), orientation="v"),
        margin=dict(b=20, l=50, r=65),
    )
    metrics = list(set(df["Metric"]))
    rows_num = len(metrics)

    if rows_num == 2:
        row_width = [0.5, 0.5]
    else:
        row_width = [1]

    fig = make_subplots(
        rows=rows_num,
        cols=1,
        shared_xaxes=False if len(list(set(df["default_period_type"]))) > 1 else True,
        vertical_spacing=vertical_spacing,
        row_width=row_width,
    )

    for metric, metric_df in df.groupby("Metric", sort=False):
        metric_idx = metrics.index(metric)
        for ticker, ticker_df in metric_df.sort_values(["PeriodDateTime"]).groupby("Ticker"):
            idx = tickers.index(ticker)
            if chart_type == "bar":
                fig.add_trace(
                    go.Bar(
                        x=ticker_df["Period"],
                        y=ticker_df["value"],
                        name=ticker,
                        marker=dict(color=px.colors.qualitative.Plotly[idx]),
                        legendgroup=f"group{idx}",
                        showlegend=True if metrics.index(metric) == 0 else False,
                    ),
                    row=metrics.index(metric) + 1,
                    col=1,
                )
            else:
                fig.add_trace(
                    go.Scatter(
                        x=ticker_df["Period"],
                        y=ticker_df["value"],
                        name=ticker,
                        marker=dict(color=px.colors.qualitative.Plotly[idx]),
                        legendgroup=f"group{idx}",
                        showlegend=True if metrics.index(metric) == 0 else False,
                    ),
                    row=metrics.index(metric) + 1,
                    col=1,
                )
        if len(metrics) > 1:
            fig["layout"][f"xaxis{metric_idx+1}"]["dtick"] = round(len(metric_df["Period"]) / 10)

    fig.update_traces(hovertemplate="%{y} %{x}")

    sorted_periods = df.sort_values("PeriodDateTime")["Period"].unique()
    sorted_periods_filtered = (
        sorted_periods[:: int(len(sorted_periods) / 10)] if len(sorted_periods) > 10 else sorted_periods
    )  # Get only 10,

    if "fin_metric" in set(metrics_type_df["type"]):
        # Reverse the financials periods. e.g: `FY 2020` will converted to `2020 FY``
        sorted_periods = list(map(reverse_period, sorted_periods))

    fig.update_xaxes(
        rangeslider_visible=False,
        spikemode="across",
        spikesnap="cursor",
        categoryorder="array",
        categoryarray=sorted_periods,
        xaxis_tickvals=sorted_periods_filtered,
        xaxis_ticktext=sorted_periods_filtered,
    )

    fig.update_layout(
        **layout,
        **get_plot_default_layout(),
        legend_title_text=None,
        xaxis_title=None,
        yaxis_title=None,
    )

    # Add yaxis titles
    for idx, r in enumerate(np.cumsum(row_width)[::-1]):
        fig["layout"][f"yaxis{idx+1}"]["title"] = metrics[idx]

    fig.update_annotations(
        dict(
            font_size=10,
            font_color="#525252",
        )
    )
    fig.update_yaxes(
        title_font_size=10,
    )

    return fig


def historical_metrics_df2tree(df: DataFrame) -> Tree:
    df = df.pivot(index=["Metric", "Ticker"], columns="Period", values="value").reset_index()
    reversed_periods = [reverse_period(p) for p in list(df.columns[2:])]
    reversed_periods.sort(reverse=True)
    sorted_periods = [reverse_period(rp) for rp in reversed_periods]
    col_width = 15
    plot_title = f"Historical {' and '.join(list(set(df['Metric'])))}"
    tree = Tree(Panel(plot_title, width=50))
    # Add header table to tree
    header_table = Table(
        width=55 + (col_width * (len(sorted_periods) - 1)), show_lines=False, show_header=False, box=None
    )
    header_table.add_column(width=15, style="magenta")
    for p in list(df):
        header_table.add_column(width=col_width, justify="right", style="magenta")
    header_table.add_row("Period", *sorted_periods)
    tree.add(header_table)

    table_style = get_table_style("metrics_historical")

    for metric, metric_values in df.groupby("Metric", sort=False):
        metric_branch = tree.add(f"[magenta]{metric}")
        for i, r in metric_values.reset_index().iterrows():
            t = Table(
                width=22 + (col_width * (len(sorted_periods) - 1)),
                show_lines=False,
                show_header=False,
                box=None,
                row_styles=[table_style["row_styles"][i % 2]],
            )
            t.add_column(width=11)
            for period in sorted_periods:
                t.add_column(width=col_width, justify="right")
            t.add_row(r["Ticker"], *[f"{d}" if d is not np.nan else "-" for d in r[sorted_periods].values])
            metric_branch.add(t)

    return tree


@metrics.command()
@click.pass_context
@click.option(
    "--tickers",
    "-k",
    type=TickerParamType(),
    required=True,
    callback=validate_tickers,
    help="Comma-separated list of tickers.",
)
@click.option(
    "--metrics",
    "-m",
    type=MetricParamType(),
    default="pe_ratio_ttm",
    help="Comma-separated list of daily metrics.",
)
@click.option(
    "--output",
    "-o",
    type=OutputParamType(),
    default="terminal",
    help="Output can be terminal or plot.",
)
@click.option(
    "--plot_type",
    "-p",
    type=ChartParamType([("bar", "Bar chart"), ("line", "Line chart")]),
    default="line",
    help="Plot can be bar or line chart.",
)
@click.option("--pivot", "-pv", is_flag=True, default=False, help="Output will be pivot table.")
@click.option(
    "--period_type",
    "-t",
    type=PeriodTypeParamType(),
    help="Period by which you want to view the report. Possible values are `D` for daily, \
        `FY` for yearly, `Q` for quarterly, `TTM` for TTM reports, `YTD` for YTD reports.",
)
@click.option("--from_date", "-f", type=DateTime(), help="Histotical metrics from date.")
@click.option("--to_date", "-t", type=DateTime(), help="Histotical metrics to date.")
@pass_command
def historical(
    ctx: click.Context,
    tickers: str,
    metrics: str,
    output: str,
    plot_type: str,
    pivot: bool,
    period_type: Optional[str],
    from_date: Optional[datetime],
    to_date: Optional[datetime],
) -> None:
    """
    Lists, compares and plots metrics of given companies. TICKERS is a comma-separated list of tickers.

    Examples:

    `i8 metrics historical --metrics net_income --tickers AMD,INTC,QCOM --output plot --plot_type bar --period_type Q`
    `i8 metrics historical --metrics total_revenue,total_assets --tickers AMD,INTC,QCOM --output terminal \
        --period_type FY --pivot`

    """
    metrics_list = metrics.replace(" ", "").split(",")
    if output not in ["terminal", "plot"]:
        click.echo(click.style(f"`{output}` is not valid output type.", fg="yellow"))
        return
    if output == "plot" and len(metrics_list) > 2:
        click.echo(click.style("For the `plot` output type you can enter up to only two metrics.", fg="yellow"))
        return
    command_path_parsed_options_dict = {
        "--tickers": tickers,
        "--metrics": metrics,
        "--output": output,
        "--plot_type": plot_type,
    }
    if period_type:
        command_path_parsed_options_dict["--period_type"] = period_type
    if from_date:
        command_path_parsed_options_dict["--from_date"] = from_date.strftime("%Y-%m-%d")
    if to_date:
        command_path_parsed_options_dict["--to_date"] = to_date.strftime("%Y-%m-%d")
    command_path = get_click_command_path(ctx, command_path_parsed_options_dict)
    tickers_list = tickers.replace(" ", "").upper().split(",")
    if len(tickers_list) > 5:
        click.echo(click.style("You can enter up to 5 tickers.", fg="yellow"))
        return
    if plot_type not in ["bar", "line"]:
        click.echo(click.style(f"`{plot_type}` is not valid chart type.", fg="yellow"))
        return
    cmd_context = {
        "command_path": command_path,
        "tickers": tickers_list,
        "plot_type": PlotType.CHART.value,
    }

    metrics_type_df: DataFrame = get_all_metrics_type_and_data_types_df()
    metrics_type_df = metrics_type_df[metrics_type_df["metric_name"].isin(metrics_list)]

    if output == "plot" and "string" in metrics_type_df["data_format"].unique():
        click.echo(
            click.style("Metrics with type `string` cannot be plotted! Select `terminal` output instead.", fg="yellow")
        )
        return

    console = Console()
    with console.status("Fetching data...", spinner="material") as status:
        df = get_historical_metrics_df(tickers_list, metrics_list, period_type, from_date, to_date)
        df = df.sort_values(["PeriodDateTime"], ascending=False).groupby(["Ticker", "Metric", "Period"]).head(1)
        if len(df["default_period_type"].unique()) > 1:
            console.print(
                (
                    "The `period type` of the provided metrics are not compatible. Make sure the provided metrics have "
                    "the same period type. Check `metrics describe` command to find more about metrics."
                ),
                style="yellow",
            )
            return
        if output == "plot":
            cmd_context["plot_title"] = f"Historical {' and '.join(list(set(df['Metric'])))}"
            status.update("Generating plot...")
            fig = create_fig(df, cmd_context, tickers_list, plot_type, metrics_type_df)

    if output == "plot":
        serve_plot(fig, cmd_context)
        return
    formatted_df = format_metrics_df(df, "console")
    if pivot:
        tree = historical_metrics_df2tree(formatted_df)
        console.print(tree)
        return
    columns_justify: Dict[str, Any] = {}
    for metric_display_name, metric_df in df.groupby("Metric"):
        columns_justify[metric_display_name] = "left" if metric_df["display_format"].values[0] == "str" else "right"
    formatted_df = formatted_df.pivot(index=["Ticker", "Period"], columns="Metric", values="value").reset_index()
    formatted_df["reversed_period"] = formatted_df.apply(
        lambda row: reverse_period(row.Period),
        axis=1,
    )
    formatted_df.sort_values(["Ticker", "reversed_period"], ascending=False, inplace=True)
    formatted_df.drop(columns=["reversed_period"], inplace=True)
    table = df2Table(
        formatted_df,
        columns_justify=columns_justify,
    )
    console.print(table)



import codecs
import os
from typing import Any

"""These methods are used in `setup.py`, so keep in mind to use only built-in packages in here."""


def read(rel_path: str) -> str:
    """
    Read a file.
    """
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), "r") as fp:
        return fp.read()


def get_version() -> Any:
    """
    Read version from a file.
    """
    for line in read("version.txt").splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")



from typing import Any, List, Optional

import plotly.express as px
from pandas import DataFrame

from i8_terminal.common.formatting import color
from i8_terminal.service_result.columns_context import ColumnsContext
from i8_terminal.service_result.service_result import ServiceResult


class EarningsListResult(ServiceResult):
    def __init__(self, data: DataFrame, columns_context: ColumnsContext):
        super().__init__(data, columns_context)

    def __repr__(self) -> str:
        return repr(self._df.head(2))

    def __create_plot_traces(self, df: DataFrame, column: str, beat_color: str) -> List[Any]:
        fig_traces = []

        fig = px.bar(
            df,
            x="period",
            y=column,
            category_orders={"period": df["period"]},
            color=beat_color,
            color_discrete_map={"Yes": color.i8_green.value, "No": color.i8_red.value},
            labels={"eps_beat_?": "Beat?", "eps_actual": "EPS $", "period": ""},
        )

        fig.update_layout(legend_title_text="EPS Beat?")
        fig.update_layout(title="EPS per Quarter", font=dict(color=color.i8_dark.value))
        for trace in range(len(fig["data"])):
            fig_traces.append(fig["data"][trace])
        return fig_traces

    def __add_traces_to_fig(self, traces: List[Any], fig: Any, row: int, col: int) -> Any:
        for trace in traces:
            fig.append_trace(trace, row=row, col=col)
        return fig

    def to_plot(self, x: str = None, y: List[str] = None, kind: str = "bar") -> Any:
        if x is None:
            x = "period"
        if y is None:
            y = ["eps_consensus", "eps_actual"]
        return self._to_plot(x, y, kind)

    def to_xlsx(self, path: str, formatter: Optional[str] = None, styler: Optional[str] = None) -> None:
        df = self.to_df()
        df.to_excel(path)



from typing import Any, Callable, Dict, List, Optional, Union

import numpy as np
import pandas as pd
import plotly.express as px
from pandas import DataFrame
from rich.console import Console
from rich.table import Table

from i8_terminal.app.layout import get_plot_default_layout
from i8_terminal.common.formatting import format_date, format_number_v2
from i8_terminal.common.layout import format_df
from i8_terminal.common.utils import concat_and
from i8_terminal.config import get_table_style
from i8_terminal.i8_exception import I8Exception
from i8_terminal.service_result.column_info import ColumnInfo
from i8_terminal.service_result.columns_context import ColumnsContext


class ServiceResult:
    def __init__(self, df: DataFrame, cols_context: ColumnsContext):
        self._df = df
        self._cols_context = cols_context

    def to_df(self, format: str = "default") -> DataFrame:
        """
        Args:
            formating: possible options are
                `raw`: no formating,
                `default`: number (e.g. 123456.7890 => 123,456.79) and display formatting (net_income => Net Income)
                `humanize`: numbers are formatted to human-friendly formats (e.g. 1200000 => $1.20 M)
        """
        return self._format_df(self._df.copy(), format)

    def _wide_df(self, format: str) -> DataFrame:
        df_formatted = self._format_df(self._df.copy(), format)
        df_raw = self._df.copy()

        return pd.concat([df_formatted, df_raw], axis=1)

    def to_json(self) -> Any:
        pass

    def to_console(self, format: str = "default", style_profile: str = "default", width: int = 800) -> None:
        table = self._to_rich_table(format, style_profile)
        console = Console(force_jupyter=True, width=width)
        console.print(table)

    def _to_rich_table(self, format: str, style_profile: str) -> Table:
        style = get_table_style(style_profile)
        table = Table(**style)
        df = self._wide_df(format)
        ci_list = self._cols_context.get_col_infos()
        non_num_dts = ["str", "string", "datetime"]

        for ci in ci_list:
            table.add_column(ci.display_name, justify="left" if ci.data_type in non_num_dts else "right")

        def _process_value(raw: Union[int, float], formatted: str, ci: ColumnInfo) -> str:
            if raw is np.nan or raw is None:
                return "-"
            value = raw if format == "raw" else formatted
            if ci.colorable and ci.data_type not in non_num_dts and raw != 0:
                color = "green" if raw > 0 else "red"
                return f"[{color}]{value}[/{color}]"
            return str(value)

        for _, r in df.iterrows():
            row = [_process_value(r[ci.name], r[ci.display_name], ci) for ci in ci_list]
            table.add_row(*row)

        return table

    def to_plot(self, x: str, y: List[str], kind: str = "bar") -> Any:
        return self._to_plot(x, y, kind)

    def _to_plot(self, x: str, y: List[str], kind: str = "bar") -> Any:
        df = self._df[[x] + y]
        df_grouped = df.groupby(x)[y].mean().reset_index()

        cid = self._cols_context.get_col_info_dict()
        y_display_names = [cid[ci_name].display_name for ci_name in y]
        title = f"{concat_and(y_display_names)} by {cid[x].display_name}"  # type: ignore

        fig = None
        if kind == "bar":
            fig = px.bar(
                df_grouped,
                hover_data=[],
                x=x,
                y=y,
                barmode="group",
                title=title,
                labels={"value": "Metric Value", x: "Period"},
            )
        if not fig:
            return

        fig.update_traces(width=0.3, hovertemplate="%{y}%{_xother}")
        fig.update_layout(
            **get_plot_default_layout(),
            bargap=0.4,
            legend_title_text=None,
            xaxis_title=None,
            margin=dict(b=15, l=70, r=20),
        )
        fig.show()

        return fig

    def to_xlsx(self, path: str, formatter: Optional[str] = None, styler: Optional[str] = None) -> Any:
        pass

    def to_csv(self, path: str, format: str = "raw") -> None:
        df = self._df.copy()
        df = self._format_df(df, format)
        df.to_csv(path, index=False)

    def _format_df(self, df: DataFrame, format: str = "default") -> DataFrame:
        ci_list = self._cols_context.get_col_infos()
        display_names: Dict[str, str] = {}
        formatters: Dict[str, Any] = {}
        for ci in ci_list:
            if ci.display_name is None or ci.data_type is None or ci.unit is None:
                raise I8Exception(f"Missing required metadata fields on colum: `{ci.name}`")

            display_names[ci.name] = ci.display_name
            if ci.data_type in ["int", "unsigned_int", "float", "unsigned_float"] and self._df[ci.name].max() < 1e6:
                if format == "raw":
                    formatters[ci.name] = self._get_formatter(ci.unit, ci.data_type, format)
                else:
                    formatters[ci.name] = self._get_formatter(ci.unit, ci.data_type, format="default")
            else:
                formatters[ci.name] = self._get_formatter(ci.unit, ci.data_type, format)
        return format_df(df, display_names, formatters)

    def _get_formatter(self, unit: str, data_type: str, format: str) -> Callable[[Any], Optional[Union[str, int, Any]]]:
        if format == "raw":
            if unit == "datetime" and data_type == "datetime":
                return lambda x: format_date(x)  # TODO: Implement a new format_date function with date format
            else:
                return lambda x: x

        if data_type == "str" or unit == "string":
            return lambda x: x

        if unit == "datetime":
            if data_type == "datetime":
                return lambda x: format_date(x)
            else:
                return lambda x: x

        if format == "default":
            if data_type in ["int", "unsigned_int"]:
                return lambda x: format_number_v2(x, percision=0, unit=unit)
            elif data_type in ["float", "unsigned_float"]:
                return lambda x: format_number_v2(x, percision=2, unit=unit)
        elif format == "humanize":
            return lambda x: format_number_v2(x, percision=2, unit=unit, humanize=True)
        elif format == "millionize":
            if data_type in ["int", "unsigned_int"]:
                return lambda x: format_number_v2(x, percision=0, unit=unit, in_millions=True)
            elif data_type in ["float", "unsigned_float"]:
                return lambda x: format_number_v2(x, percision=2, unit=unit, in_millions=True)

        return lambda x: x



import os

import click
import nbformat as nbf
from rich.console import Console

from i8_terminal.commands.notebook import notebook
from i8_terminal.common.cli import pass_command


@notebook.command()
@click.option("--name", "-n", help="Notebook file name.")
@pass_command
def launch(name: str) -> None:
    """
    Launches jupyter notebook.

    Examples:

    `i8 notebook launch --name mynotebook`

    """
    console = Console()
    with console.status("Launching notebook...", spinner="material"):
        nb = nbf.v4.new_notebook()
        text = "Welcome to notebook of i8 Terminal."

        code = (
            "from i8_terminal.config import init_notebook\n" "from i8_terminal import api as i8\n\n" "init_notebook()"
        )

        nb["cells"] = [nbf.v4.new_markdown_cell(text), nbf.v4.new_code_cell(code)]
        fname = f"{name}.ipynb"

        with open(fname, "w") as f:
            nbf.write(nb, f)

        os.system(f"jupyter notebook {name}.ipynb")



import click
from rich.console import Console
from rich.table import Table

from i8_terminal.commands.metrics import metrics
from i8_terminal.common.cli import pass_command
from i8_terminal.common.formatting import styling_markdown_text
from i8_terminal.common.metrics import get_metric_info
from i8_terminal.types.metric_param_type import MetricParamType


def get_metric_info_table(metric: str) -> Table:
    h_color = "cyan"
    metric_info = get_metric_info(metric)
    metric_info_table = Table(padding=(0, 0, 1, 1), box=None)
    metric_info_table.add_column(justify="right", width=25)
    metric_info_table.add_column(justify="left")
    metric_info_table.add_row(f"[{h_color}]Metric Name[/{h_color}]", metric)
    metric_info_table.add_row(f"[{h_color}]Display Name[/{h_color}]", metric_info["display_name"])
    metric_info_table.add_row(f"[{h_color}]Unit[/{h_color}]", metric_info["unit"])
    metric_info_table.add_row(f"[{h_color}]Type[/{h_color}]", metric_info["type"])
    metric_info_table.add_row(f"[{h_color}]Display Format[/{h_color}]", metric_info["display_format"])
    metric_info_table.add_row(f"[{h_color}]Default Period type[/{h_color}]", metric_info["default_period_type"])
    metric_info_table.add_row(
        f"[{h_color}]Url[/{h_color}]", f"[blue]https://docs.i8terminal.io/metrics/{metric}/[/blue]"
    )
    metric_info_table.add_row(f"[{h_color}]Description[/{h_color}]", styling_markdown_text(metric_info["description"]))
    if metric_info["remarks"]:
        metric_info_table.add_row(f"[{h_color}]Remarks[/{h_color}]", styling_markdown_text(metric_info["remarks"]))
    return metric_info_table


@metrics.command()
@click.option("--name", "-n", type=MetricParamType(), required=True, help="Metric name.")
@pass_command
def describe(name: str) -> None:
    """
    Describe the metric information.

    Examples:

    `i8 metrics describe --name basic_eps`
    """
    console = Console()
    metric_info = get_metric_info_table(name)
    console.print(metric_info)
